"""A request was made to drive the driver outside its configured limits."""


class DriverLimitsError(AssertionError):

    """A request was made to drive the driver outside its configured limits."""
